#include "../rw.h"

#ifndef DFF_H
#define DFF_H

#define CLUMP 0x10
#define FRAMELIST 0xE
#define GEOMETRYLIST 0x1A
#define GEOMETRY 0xF
#define MATERIALLIST 0x8
#define MATERIAL 0x7
#define TEXTURE 0x6
#define ATOMIC 0x14

#define STRUCT 0x1
#define STRING 0x2
#define EXTENSION 0x3

#define FRAME 0x253F2FE
#define HANIM 0x11E
#define BINMESH 0x50E
#define NATIVEDATA 0x510
#define SKIN 0x116
#define PARTICLES 0x118
#define MORPH 0x105
#define SKYMIPMAP 0x110
#define MATERIALEFFECTS 0x120
#define RIGHTTORENDER 0x1F

#define SPECULARMAT 0x253F2F6
#define REFLECTIONMAT 0x253F2FC
#define MESHEXTENSION 0x253F2FD
#define COLLISIONMODEL 0x253F2FA
#define NIGHTVERTEXCOLOR 0x253F2F9
#define ADCPLG 0x134

#define PS2 1
#define PC  2     

#define X 0              
#define Y 1         
#define Z 2             

#define U 0        
#define V 1        

#define R 0       
#define G 1       
#define B 2          
#define A 3

int platform;

typedef struct atomic {
	int frameindex;
	int geometryindex;
	int unknown0;
	int unknown1;
} atomic;

typedef struct split {
	int indexcount;
	int vertexcount;
	int matindex;

	/* PC only */
	int *indices;

	/* These are used by VC PS2. PC data is converted to fit PS2 format */
	float **vertices;
	float **uv[2];
	unsigned char **vc;
	float **normals;

	/* SA PS2 only */
	int saindexcount;
	float **savertices;
	float **sauv[2];
	float **sanormals;
	unsigned char **savc;
	int *saflags;
	int savertexcount;
} split;

typedef struct texture {
	short filterflags;
	short unknown;
	char *tex;
	char *alpha;
} texture;

typedef struct material {
	int unknown0;
	char color[4];
	int unknown1;
	int texcount;
	float unknown2;
	float unknown3;
	float unknown4;
	int tc;
	texture *tex;
} material;

typedef struct geometry {
	short flags;
	unsigned char uvsets;
	unsigned char unknown;
	unsigned char newuvsets;
	int facec;
	int vertexc;
	int framec;
	float ambient;
	float diffuse;
	float specular;

	float bs[4];
	int bspos;
	int bsnor;

	/* PC only, partially used
	 * Vertices are converted to VC PS2 format */
	unsigned char **vc;
	float **uv[2];
	unsigned short **faces;
	float **vertices;
	float **normals;

	int matcount;
	int mc;
	material *mat;

	int facetype;
	int splitcount;
	int indexcount;
	split *splt;
} geometry;

typedef struct frame {
	float rotmatrix[4][4];
	float posmatrix[4][4];
	float absmat[4][4];

	int parent;
	int unknown;

	char *name;

	int display;
} frame;

typedef struct clump {
	int objcount;
	int oc;
	atomic *atm;
	int frmcount;
	int fc;
	frame *frm;
	int geocount;
	int gc;
	geometry *geo;
} clump;

extern void ReadDff(clump *clp, FILE *dff);
extern void WriteObj(clump *clp, FILE *obj);
extern void WriteDff(clump *clp, FILE *newdff);
extern void WriteSAObj(clump *clp, FILE *obj);
extern void ReadBinMeshPS2(geometry *geo, FILE *dff);
extern void ReadBinMeshPC(geometry *geo, FILE *dff);
extern void ReadNativeDataVCPS2(geometry *geo, FILE *dff);
extern void ReadNativeDataSAPS2(geometry *geo, FILE *dff);

#endif
